SUMMARY
=======

TOSIBOX(R) Lock for Container is a VPN server software. You can establish a
secure VPN connectivity from TOSIBOX(R) Key to the host running the Lock, or
to the same LAN the Lock is running on.

Lock for Container can be used in different ways. These require a bit different
configurations, because of the way Docker handles the networks.

1. access the host (or other container) running Lock
2. access external LAN (virtual/physical) of the host running Lock
3. same as 1, but through TOSIBOX(R) Hub
4. same as 2, but through TOSIBOX(R) Hub
5. establish connectivity from host running Lock to LAN of TOSIBOX(R) Hub
6. establish connectivity from the LAN of host running Lock to LAN of TOSIBOX(R) Hub
7. access LAN side of the gateway running Lock
8. same as 7, but through TOSIBOX(R) Hub
9. establish connectivity from the LAN of gateway running Lock to LAN of TOSIBOX(R) Hub

INSTALL WIZARD
==============

To install Lock for Container in wizard mode, just enter:

`# tosibox-lock-for-container.sh install`

This is an automated process, which asks you few questions and does all of the
required installation steps for you.


INSTALLATION
============

In order to install the lock for container, you need to install the software.

`# tosibox-lock-for-container.sh install --mac 123456000000 --port 8000`

This will install the Lock image to the docker and run the container. It will
publish the maintenance interface on http://localhost:8000
It needs to be activated to be run.

ACTIVATION
==========

To start using the Lock, you need to have a license for activation. You may
order a license from https://www.tosibox.com/activate

`# tosibox-lock-for-container.sh activate --license XXXXXXXXXXXXXXX`

You need to accept the EULA before activating the Lock.

ROUTING
=======

In all use cases, you need to create a static route to the hosts you want to
access. By default this will be done to the default address space 172.17.0.1.

`# tosibox-lock-for-container.sh route --network 172.17.0.0/24 \
                                       --netmask 255.255.255.0 \
                                       --gateway 172.17.0.1`

If you want to create a route to your local LAN, just do the same with your LAN
parameters (in this example 192.168.1.0/24):

`# tosibox-lock-for-container.sh route --network 192.168.1.0/24 \
                                       --netmask 255.255.255.0 \
                                       --gateway 192.168.1.1 `

NETWORKING
==========

If you want to create your own network, there are options: host, internal or
external.

host
----
Not supported at the moment. This should bind container directly to the hosts
networking, but for some reason it does not currently work.

internal
--------
Bridged network creates a NAT network with your preferred parameters.

`# tosibox-lock-for-container.sh net --type internal --subnet 10.10.10.0/24`

In this case the host/gateway will get address 10.10.10.1, and the lock
will get 10.10.10.2. So you may access the lock admin interface on:
http://10.10.10.2/ on the host.

external
--------
Not supported at the moment. Macvlan can be tied to a physical network interface
of the host in the same IP address range as your local physical LAN. This means
that static addresses are reserved for the docker on this LAN. Ensure you have
these addresses free. One static address is required for the host, one for
the gateway, and one IP for the Lock.

`# tosibox-lock-for-container.sh net --type internal \
                                     --subnet 10.10.10.100/24 \
                                     --langw 10.10.10.100 \
                                     --interface eth0 \
                                     --lock 10.10.10.101 \
                                     --host 10.10.10.102`

MATCHING
========

To enable access to the Lock for a TOSIBOX(R) Key, you may do it by remote
matching in to the Key.

To start the remote matching and get code, run the following command, and
insert the code to your Key software.

`# tosibox-lock-for-container.sh match --start`
