#!/bin/bash
# Lock for Container installer
# Copyright © 2020 Tosibox Oy. All rights reserved.

LICENSE_VALID=0
VERSION="0.9.2"

echo "TOSIBOX® Lock for Container. Copyright© 2020 Tosibox Oy. All rights reserved."

# check that docker is installed and running
command -v docker >/dev/null 2>&1 || { echo >&2 "Docker not installed. Install first Docker by running e.g. 'apt install docker.io'. Aborting."; exit 1; }
docker ps >/dev/null 2>&1 || { echo >&2 "Docker engine not running. Aborting."; exit 1; }

[ -w . ] || { echo >&2 "Write permissions denied for the current directory. Aborting."; exit 1; }

HELPOPTIONS="
Make sure you have installed a recent version of Docker prior to installing Lock for Container.
To install Lock for Container copy this file on to the device and execute it with root or sudo privileges.
For more information see http://www.tosibox.com/LockForContainer

Installation:   tosibox-lock-for-container-installer.sh install
More help:      tosibox-lock-for-container-installer.sh --help
"

OPTIONS="
usage: tosibox-lock-for-container-installer.sh [COMMAND] [OPTIONS]
Commands:
install              Install Lock for Container on Docker (with defaults)
activate             Activate Lock with license/activation code
uninstall            Uninstall Lock for Container on Docker
update               Update Lock for Container
match                Matching enablement/disablement/resetting for Keys
start                Start Lock for Container (if stopped)
stop                 Stop Lock for Container
restart              Restart Lock for Container

Installation options: install
-m, --hostname [hostname]           Hostname of the Lock device
-p, --port [port]                   Lock admin web interface port exposed to host (not needed for external lan type) Default: 8000
--namespace                         Namespace for identifying multiple Lock for Container instances on the same host

Update options: update
-r [version]                        Version to be updated to

Activation options: activate
-l, --license [activation code]     Activation code from Tosibox

Matching options: match
--start                             Generate Remote Matching Code (use code with TOSIBOX(r) Key)
--cancel                            Cancel Remote Matching Code (invalidate the generated code)
--reset                             Reset all matched TOSIBOX(r) Keys

Import/export: backup
--import-config                     Import the Lock configs / identity to a new LfC
--export-config                     Export the Lock configs / identity to a file

Generic options:
-h, --help                          Show this help
-v, --version                       Show version
--dryrun                            Dry run (only echo the commands to be run)
"

OPTIONS_EXPERIMENTAL="
usage: tosibox-lock-for-container-installer.sh [COMMAND] [OPTIONS]
Commands:
install              Install Lock for Container on Docker
activate             Activate Lock with license/activation code
uninstall            Uninstall Lock for Container on Docker
update               Update Lock for Container
route                Add static route to LAN
net                  Configure network / subnet for the container
match                Matching enablement/disablement/resetting for Keys
backup               Backup / recover container
start                Start Lock for Container (if stopped)
stop                 Stop Lock for Container
restart              Restart Lock for Container
shell                Open shell on the Container

Installation options: install
-m, --hostname [hostname]           Hostname of the Lock device
-p, --port [port]                   Lock admin web interface port exposed to host (not needed for external lan type) Default: 8000
--namespace                         Namespace for identifying multiple Lock for Container instances on the same host

Update options: update
-r [version]                        Version to be updated to

Activation options: activate
-l, --license [activation code]     Activation code from Tosibox

Routing specific in the Lock for Container: route
-n, --network [network]             Network address for static routing the container traffic e.g. 192.168.1.0/24
-a, --netmask [netmask]             Netmask for the static route e.g. 255.255.255.0
-g, --gateway [gateway]             Gateway IP address e.g. 10.10.10.1

Network specific options: net
-y, --type [type]                   LAN Network setup type [default/host/internal/external].
-s, --subnet [subnet]               (internal/external) macvlan/bridge subnet address e.g. 10.10.10.0/24
-w, --langw [lan gateway]           (external) LAN gateway e.g. 10.10.10.100
-i, --interface [interface]         (external) interface to link the macvlan with
-t, --host                          (external) host address in the subnet e.g. 10.10.10.102
-k, --lock                          (external) lock address in the subnet e.g. 10.10.10.101
--persist                           Persist packet forwarding rules on host (experimental)

Matching options: match
--start                             Generate Remote Matching Code (use code with TOSIBOX(r) Key)
--cancel                            Cancel Remote Matching Code (invalidate the generated code)
--reset                             Reset all matched TOSIBOX(r) Keys

Import/export: backup
--import-config                     Import the Lock configs / identity to a new LfC
--export-config                     Export the Lock configs / identity to a file
-b, --import [filename]             Import container from backup file
-x, --export [filename]             Export container to backup file, timestamp will be added

Generic options:
-h, --help                          Show this help
-v, --version                       Show version
--dryrun                            Dry run (only echo the commands to be run)
"

#Routing specific the Host: hostroute
# these settings are needed for communicating through Virtual Central Lock to other locks and devices behind the TOSIBOX(R) VCL
# EXPERIMENTAL, NOT OFFICIALLY SUPPORTED IN LfC 1.0.0
#--hr-network [network]             Network address for routing the host traffic through Lock for Container e.g. 192.168.1.0/24
#--hr-netmask [netmask]             Netmask for the route e.g. 255.255.255.0 (not needed if CIDR prefix is given. This overrides CIDR prefix, if given.)
#--hr-gateway [gateway]             Gateway IP address for the route e.g. 10.10.10.1

# DEFAULTS
# registry address for fetching container image
REGISTRY_HOST="https://eu.gcr.io"
LAN_NW_NAME="tbnet"
CONTAINER_NAME="lock-container"
LAN_LINK_NAME="tblink"
REGISTRY="eu.gcr.io/tosibox-oy/softx-openwrt"
PRODUCT="Lock for Container"
VARIANT="latest"
LAN_TYPE="internal"  # use bridge by default
HOSTNAME="$CONTAINER_NAME"
PORT="8000"
NAMESPACE=""
DR="" # by default no dry run
CONTAINER_IP_FILE="lock-for-container-ip"
DEFAULT_PASSWORD="admin"
CONFIG_FILENAME=""
WIZARD=0
# LOGGING
RED='\033[0;31m'
NC='\033[0m' # No Color
DRYRUN_ECHO="echo -e ${RED}"
LOG_ECHO="echo -e ${NC}"

MD5="echo md5 not found, can't verify: "
[ -x "$(command -v md5)" ] && MD5=md5         # posix
[ -x "$(command -v md5sum)" ] && MD5=md5sum   # MacOS (older)

PARAMCOUNT="$#"

# shared key for downloading the Docker image from the registry
# TODO: move to public Docker Hub registry

KEY='{
  "type": "service_account",
  "project_id": "tosibox-oy",
  "private_key_id": "46904dc689f2458b8e0986415ca20b00d3450ed1",
  "private_key": "-----BEGIN PRIVATE KEY-----\nMIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQDBl3AUXI+xzYcZ\nWlop7iRUKywgmKvPe/hC/NKa8dU/9okHi7yh7oLywQvNTsC7VyP55gknDR1TTqim\nkpMi7wtQRpWS4DcoWKLMv2IIM9wgmQR0ywd7J5SiJ+MVkpT9Q5mG00pDJkWoRFoD\n7Al1pvSM9E1Tm3aaDEy5VYQPF9cwvzTSh2A9a5mVYq4IrTYqVopkQL6B3wGmXOZS\nGuO/lvS/YBcaTPg+r80ozI7evUoGB3ElpteTpgjsdXmBHDrxWFlALliOYHk8bgHW\nCLHBb3I5Lx1YNIeOcKBr2RkDLeIk6vkxFtFLoIdfPJo9Qi1HCUcccHGhIECrQ87H\n5Xx377O9AgMBAAECggEAJ/HJ2AGzSpIoXpRUh1NSXqWb0Mv/o2umcZf+Eqd0g2Ko\nqoxsKxNuJI2tlDq8LMuRqkYLPGoJ038m8FvNDUe6k2FFCo3GA8aqLMoOfbfJo1Y/\nY6Pi0sf7dKv7QJ/CqzFQE9A0Af/7HnmgAzAoUYTV7+GWtPoecjG4gQNCoOSf67gk\nVlhbrqh3Y7xVvSHCRQe7Wk2I7Y6EbCRN0gcrNA1/FoHi38yVxmjekXc8K7chmt0V\n+HrP+Zz2YGv/SoimOFhLVUOHkgKUVSmqzw9MpOsnIEQy6DKGVQ0TdjzvwBdzNiyy\nrhkaBeVmClUTBIcBGhsHlZFG3EMHkR72XeJe5l9S5wKBgQDj/RFHyYT/EI7McJn1\nCJshM/wizCM87NVcE2OnsXuPvro6zsmrLNqUU5HiHdIgTOtuwgOxuYSUroYtZVEu\nuQggzqzrMjFKQlbT26Pvbsp3QkL5zB+XZPuSPEX0+OMp3rwYFd6qvqQxf5HPb2P0\nlVJXFEZ56TnJ412EDmptuarAawKBgQDZYHr9H/wIdY4jbbwzP+PIyqCcS4d/zUog\nkrIZjGUI+hRFUD8b1EPlPCTh1pRzeHIQsM5LHEmlSTOKfxbrhomlQpFXmQISq3zB\nshLUXD8cDHg8y1Bh1/f+iHgY9JB1F/uT9ImBi2p17f7+E/YQKPOb9KX11X38GjtS\nkxLmq5lGdwKBgQChelLM96ydQ8sAfhhtdyT5BWrl8pRayJvt1UXWNlz/2OTqViB8\nfw5p1C8Vam2ztwdhsUUQkgeXj201pfc0EpXxNrE+/JPwtr/s2jpokZ4RjfsGFk+h\n2UJdtxcNhWLP+xRaMtM9OD3vWR2bQJgZWdGyH4gzPL9TgRhxOgyjnseT9QKBgC70\nFBzQ59O78LPeZCR9zGn2urKH7gDBCx32EdpK5RgHTJGcP8V3GCNP1nehrfczMRBB\n4I2A0reNFED6Kq09JkY4FsThG/2EZTJBkKemNUuNtfsLv5Ui3UJxn99kFD/2qjTP\nQVRmFNzcR5QCI8hh0B0Dv0VWlBv8Xov7kL2am0UJAoGAHl2zdLqSpmfEOku9wbim\nfaU4odoRNZ01fjqLXAlKxB4C9cssVNPeqAy2y4QPrYdjuO0HjQp49/rHux6zD4oE\nerGQ/njBRwZvhbHuepc4owb2VgEkXowB3EH6VgsRkUOPcU+4tlv7kgLK4CdULJ0w\nSatJzLHj6YqfkET+8UdsPlo=\n-----END PRIVATE KEY-----\n",
  "client_email": "softx-image-puller@tosibox-oy.iam.gserviceaccount.com",
  "client_id": "108813524093408648050",
  "auth_uri": "https://accounts.google.com/o/oauth2/auth",
  "token_uri": "https://oauth2.googleapis.com/token",
  "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
  "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/softx-image-puller%40tosibox-oy.iam.gserviceaccount.com"
}'


################################################################################
# ACTIVATION
################################################################################

activate() {
  log "Activating Lock for Container. This will take a while.."
  docker_exec "uci set tosibox.settings.eula_accepted='1'"
  # Add Google name server in order the default doesn't work
  docker_exec "echo 'nameserver 8.8.8.8' >> /etc/resolv.conf"
  docker_exec "env LICENSE_KEY=$LICENSE /usr/share/tosibox/create_license.sh &> /dev/null"
  docker_exec "/usr/share/tosibox/activate_license &> /dev/null"
  # Restart tbstatus to make sure interprocess communication works
  docker_exec "/etc/init.d/tbstatus restart &> /dev/null"
  log "Lock for Container activated."
  log
}

activation_help() {
  log "Activation aborted. You can do this later from the web use interface by clicking"
  log "'Activate license' below Internet status or from command line by running this script with the activate switch."
  log
}

ask_activation_license() {
  log "Please insert valid Activation Code or break by pressing return:"
  read -p "" LICENSE
  while true; do
    if [ -z "$LICENSE" ]; then
      activation_help
      ACTIVATE=""
      break;
    fi
    validate_license $LICENSE
    LICENSE_VALID=$?
    if [ "$LICENSE_VALID" -eq 1 ]; then
      log "OK"
      break;
    else
      log "Invalid license, please try again or break by pressing return:"
      read -p "" LICENSE
    fi
  done
}

validate_license() {
  license="$1"
  if [ ${#license} -ne 25 ]; then
    return 0
  fi
  version=$(echo $license | cut -c1-1)
  id=$(echo $license | cut -c2-9)
  passwd=$(echo $license | cut -c10-24)
  checksum=$(echo $license | cut -c25-25)
  calculated=$(echo -n $version$id$passwd | $MD5 | cut -c1-1)

  if [ $calculated != $checksum ]; then
    return 0
  else
    return 1
  fi
}

log() {
  $LOG_ECHO"$@"
}

################################################################################
# ROUTING / NETWORKING
################################################################################

add_static_route() {
  log "Adding static route for the given network"
  # set static route
  docker_exec "/sbin/uci add network route"
  docker_exec "/sbin/uci set network.@route[-1].interface=lan"
  docker_exec "/sbin/uci set network.@route[-1].target="$NETWORK
  docker_exec "/sbin/uci set network.@route[-1].netmask="$NETMASK
  log "Commiting network settings.."
  docker_exec "/sbin/uci commit network"
  # set to sublock mode
  docker_exec "uci set tosibox.settings.no_client=0"
  docker_exec "uci commit tosibox"
  docker_exec "pkill -HUP -x tosiboxd"
}

# not tested
add_hostroute() {
  # enable routing on the LfC
  docker_exec "iptables -t nat -A POSTROUTING -o tun0 -j MASQUERADE"

  log "Adding route from host to the given network via the Lock for Container. Needs sudo privileges"
  if [ $HOSTROUTE_NETWORK ] && [ $HOSTROUTE_GATEWAY ] && [ $HOSTROUTE_NETMASK ]; then
    $DR $sudo route add -net $HOSTROUTE_NETWORK netmask $HOSTROUTE_NETMASK gw $HOSTROUTE_GATEWAY
  else
    $DR $sudo ip route add $HOSTROUTE_NETWORK_CIDR via $HOSTROUTE_GATEWAY
  fi

  log "Setting iptables for route.."
  $DR $sudo iptables -A FORWARD -i eth0 -j ACCEPT
  log "Hostroutes added."
}

add_network_config() {
  log "Configuring IP packet forwarding on the Host to allow routing..."

  while true; do
    read -p "Do you want to configure packet forwarding on the host (Linux only, needs sudo privileges) (y/n) " yn
    case $yn in
        [Yy]* )
          $DR $sudo /sbin/sysctl -w net.ipv4.conf.all.forwarding=1
          $DR $sudo /sbin/iptables -P FORWARD ACCEPT
          break;
          ;;
        [Nn]* ) break;;
        * ) log "Please answer yes or no.";;
    esac
  done

  if [ $PERSIST_NETCONFIG ]; then

    # If these don't work, you may use other ways of persisting after the proper configuration is in place:
    # iptables-save > /etc/tosibox-routi  ng.rules
    # echo 'iptables-restore < /etc/tosibox-routing.rules' | sudo tee /etc/networkd-dispatcher/routable.d/50-iptables-restore
    log "This is still an experimental way to add persistence, please ensure that your config files are propely configured after this operation."
    log "It will add configurations to the end of /etc/network/interfaces and /etc/sysctl.conf files."

    CONT=1
    while true; do
      read -p "Do you want to continue (y/n) " yn
      case $yn in
          [Yy]* ) break;;
          [Nn]* ) CONT=0;break;;
          * ) log "Please answer yes or no.";;
      esac
    done

    if [ $CONT ]; then
      $DR echo "
# added by Tosibox lock-cli.sh for persistent packet forwarding for the Lock for Container. Ensure the config is correct.
auto $LAN_INTERFACE
iface $LAN_INTERFACE inet dhcp
post-up /sbin/iptables -P FORWARD ACCEPT
post-up /sbin/sysctl net.ipv4.conf.all.forwarding=1
# end Tosibox addition
" >> ~/etc/network/interfaces

      $DR echo "
# added by Tosibox lock-cli.sh for persistent packet forwarding
net.ipv4.ip_forward=1
# end Tosibox addition
" >> ~/etc/sysctl.conf
    fi
  fi

  # Create a docker network with the selected network driver
  # Also erase the old one if found
  $DR docker network disconnect $LAN_NW_NAME $CONTAINER_NAME &> /dev/null
  $DR docker network rm $LAN_NW_NAME &> /dev/null
  $DR docker network disconnect bridge $CONTAINER_NAME &> /dev/null

  log "Creating docker network $LAN_NW_NAME..."
  case $LAN_TYPE in
      default) log "Using the default network..";;
      host) log "Using host network...";;
      internal) log "Creating a bridge network.."; $DR docker network create -d bridge --subnet=$LAN_SUBNET_CIDR --gateway=$LAN_GW $LAN_NW_NAME;;
      external) log "Creating a macvlan network.."; $DR docker network create -d macvlan --subnet=$LAN_SUBNET_CIDR --gateway=$GATEWAY -o parent=$LAN_INTERFACE $LAN_NW_NAME;;
      * ) log "Unknown network type";;
  esac

  $DR docker network connect $LAN_NW_NAME $CONTAINER_NAME &> /dev/null
}

# Work in progress
ask_hostroute_config() {
  if [ $DEFAULT_REMOTENETWORK ]; then
    read -p "Enter network IP address for Host to route traffic to via Lock for Container ($DEFAULT_REMOTENETWORK): " IP
  else
    read -p "Enter network IP address for Host to route traffic to via Lock for Container: " IP
  fi
  while true; do
    if valid_ip $IP; then
      HOSTROUTE_NETWORK=$IP
      break;
    elif [ -z $IP ] && [ $DEFAULT_NETWORK ]; then
      HOSTROUTE_NETWORK=$DEFAULT_NETWORK
      break;
    elif [ -z $IP ]; then
      log "Aborted setting up static route."
      break;
    elif [ -z $NETWORK ]; then
      read -p "Please enter valid ip or press return to use default ($DEFAULT_NETWORK): " IP
    else
      read -p "Please enter valid ip or press return to abort: " IP
    fi
  done
  if [ $DEFAULT_NETMASK ]; then
    read -p "Enter network netmask for Lock for Container ($DEFAULT_NETMASK): " IP
  else
    read -p "Enter network netmask: " IP
  fi
  while true; do
    if valid_ip $IP; then
      NETMASK=$IP
      add_static_route
      break;
    elif [ -z $IP ] && [ $DEFAULT_NETMASK ]; then
      NETMASK=$DEFAULT_NETMASK
      add_static_route
      break;
    elif [ -z $IP ]; then
      log "Aborted setting up static route."
      break;
    elif [ $DEFAULT_NETMASK ]; then
      read -p "Please enter valid ip or press return to use default ($DEFAULT_NETMASK): " IP
    else
      read -p "Please enter valid ip or press return to abort " IP
    fi
  done
}

# gives the CIDR prefix based on the netmask given as a parameter
IPprefix_by_netmask() {
  c=0 x=0$( printf '%o' ${1//./ } )
  while [ $x -gt 0 ]; do
     let c+=$((x%2)) 'x>>=1'
  done
  echo /$c ;
}

valid_ip() {
  local  ip=$1
  local  stat=1

  if [[ $ip =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$ ]]; then
    OIFS=$IFS
    IFS='.'
    ip=($ip)
    IFS=$OIFS
    [[ ${ip[0]} -le 255 && ${ip[1]} -le 255 \
       && ${ip[2]} -le 255 && ${ip[3]} -le 255 ]]
      stat=$?
  fi
  return $stat
}

ask_network_config() {
  if [ $DEFAULT_SUBNET ]; then
    read -p "Enter network IP address space for Container subnet ($DEFAULT_SUBNET): " IP
  else
    read -p "Enter network IP address space for Container subnet: " IP
  fi
  while true; do
    if valid_ip $IP; then
      LAN_SUBNET_CIDR=$IP
      break;
    elif [ -z $IP ] && [ $DEFAULT_SUBNET ]; then
      LAN_SUBNET_CIDR=$DEFAULT_SUBNET
      break;
    elif [ -z $IP ]; then
      log "Aborted setting up Container subnet."
      break;
    elif [ -z $LAN_SUBNET_CIDR ]; then
      read -p "Please enter valid ip or press return to use default ($DEFAULT_SUBNET): " IP
    else
      read -p "Please enter valid ip or press return to abort: " IP
    fi
  done
  if [ $DEFAULT_NETMASK ]; then
    read -p "Enter network netmask for Container subnet ($DEFAULT_NETMASK): " IP
  else
    read -p "Enter network netmask for Container subnet: " IP
  fi
  while true; do
    if valid_ip $IP; then
      NETMASK=$IP
      IPPREFIX=$(IPprefix_by_netmask $NETMASK)
      LAN_SUBNET_CIDR="$LAN_SUBNET_CIDR$IPPREFIX"
      log "The CIDR notation for the network is: $LAN_SUBNET_CIDR"
      add_network_config
      break;
    elif [ -z $IP ] && [ $DEFAULT_NETMASK ]; then
      NETMASK=$DEFAULT_NETMASK
      IPPREFIX=$(IPprefix_by_netmask $NETMASK)
      LAN_SUBNET_CIDR="$LAN_SUBNET_CIDR$IPPREFIX"
      echo "The CIDR notation for the network is: $LAN_SUBNET_CIDR"
      add_network_config
      break;
    elif [ -z $IP ]; then
      log "Aborted setting up network."
      break;
    elif [ $DEFAULT_NETMASK ]; then
      read -p "Please enter valid IP netmask address or press return to use default ($DEFAULT_NETMASK): " IP
    else
      read -p "Please enter valid IP netmask address or press return to abort " IP
    fi
  done
}

ask_route_config() {
  if [ $DEFAULT_NETWORK ]; then
    read -p "Enter network IP address for Lock for Container to route traffic to ($DEFAULT_NETWORK): " IP
  else
    read -p "Enter network IP address for Lock for Container to route traffic to: " IP
  fi
  while true; do
    if valid_ip $IP; then
      NETWORK=$IP
      break;
    elif [ -z $IP ] && [ $DEFAULT_NETWORK ]; then
      NETWORK=$DEFAULT_NETWORK
      break;
    elif [ -z $IP ]; then
      log "Aborted setting up static route."
      break;
    elif [ -z $NETWORK ]; then
      read -p "Please enter valid ip or press return to use default ($DEFAULT_NETWORK): " IP
    else
      read -p "Please enter valid ip or press return to abort: " IP
    fi
  done
  if [ $DEFAULT_NETMASK ]; then
    read -p "Enter network netmask for Lock for Container ($DEFAULT_NETMASK): " IP
  else
    read -p "Enter network netmask: " IP
  fi
  while true; do
    if valid_ip $IP; then
      NETMASK=$IP
      add_static_route
      break;
    elif [ -z $IP ] && [ $DEFAULT_NETMASK ]; then
      NETMASK=$DEFAULT_NETMASK
      add_static_route
      break;
    elif [ -z $IP ]; then
      log "Aborted setting up static route."
      break;
    elif [ $DEFAULT_NETMASK ]; then
      read -p "Please enter valid ip or press return to use default ($DEFAULT_NETMASK): " IP
    else
      read -p "Please enter valid ip or press return to abort " IP
    fi
  done
}

################################################################################
# BASIC FUNCTIONALITY
################################################################################

stop() {
  while true; do
    log
    log "This will stop Lock for Container. All VPN connections will be closed."
    log "If you are remotely connected you will be disconnected."
    log "You can start Lock for Container again with the 'start' switch."
    read -p "Do you want to stop Lock for Container? (y/n) " yn
    case $yn in
      [Yy]* ) break;;
      [Nn]* ) log "Cancelled."; exit;;
      * ) log "Please answer yes or no.";;
    esac
  done
  docker stop $CONTAINER_NAME
}

restart() {
  $DR docker restart $CONTAINER_NAME
}

wait_until_ready() {
  # wait until docker runs
  while [ "`docker inspect -f {{.State.Running}} $CONTAINER_NAME`" != "true" ]; do sleep 2; done
  # wait until uci & co are ready.
  sleep 7
}

install() {
  log "Installing Lock for Container ($VARIANT)..."

  log "Login to registry..."
  echo "$KEY" | $DR docker login -u _json_key --password-stdin $REGISTRY_HOST #&> /dev/null

  # remove the old latest image, otherwise it uses the existing, which might be earlier version
  [ $VARIANT == "latest" ] && $DR docker image rm $REGISTRY:$VARIANT 2>/dev/null

  log "Starting Lock for Container..."
  case $LAN_TYPE in
    default)  $DR docker run -d --restart=unless-stopped --name=$CONTAINER_NAME --publish=$PORT:80  --cap-add=NET_ADMIN --cap-add=SYS_TIME --cap-add=SYS_PTRACE -h "$HOSTNAME" -e LICENSE_KEY=$LICENSE $REGISTRY:$VARIANT;;
    host)     $DR docker run -d --restart=unless-stopped --name=$CONTAINER_NAME --publish=$PORT:80  --cap-add=NET_ADMIN --cap-add=SYS_TIME --cap-add=SYS_PTRACE -h "$HOSTNAME" -e LICENSE_KEY=$LICENSE $REGISTRY:$VARIANT;;
    internal) $DR docker run -d --restart=unless-stopped --name=$CONTAINER_NAME --publish=$PORT:80  --cap-add=NET_ADMIN --cap-add=SYS_TIME --cap-add=SYS_PTRACE -h "$HOSTNAME" -e LICENSE_KEY=$LICENSE $REGISTRY:$VARIANT;;
    external) $DR docker run -d --restart=unless-stopped --name=$CONTAINER_NAME --ip=$LAN_LOCK_ADDR --cap-add=NET_ADMIN --cap-add=SYS_TIME --cap-add=SYS_PTRACE -h "$HOSTNAME" -e LICENSE_KEY=$LICENSE $REGISTRY:$VARIANT;;
    * ) log "Unknown network type"; exit 1;;
  esac

  $DR docker logout $REGISTRY_HOST
  log "Installation finished"

  log "Waiting for Lock for Container to wake up..."
  $DR wait_until_ready

  docker_exec "echo '$HOSTNAME' > /etc/hostname"
  docker_exec "uci set system.@system[0].hostname='$HOSTNAME'"
  docker_exec "uci commit system"
#  docker_exec "/etc/init.d/system reload"
}

update() {

  PORT=$(docker ps | grep $CONTAINER_NAME | awk -F '->80' '{print $1}' | awk '{print $NF}')

  log "Using port: "$PORT

  while true; do
    log
    log "This will update Lock for Container to version $VARIANT."
    log "All VPN connections will be closed. If you are remotely connected you will be disconnected."
    read -p "Do you want to update Lock for Container? (y/n) " yn
    case $yn in
        [Yy]* ) break;;
        [Nn]* ) log "Cancelled."; exit;;
        * ) log "Please answer yes or no.";;
    esac
  done

  EXPORT_CONFIG="update"

  log "Exporting configuration of Lock for Container..."
  export_config

  HOSTNAME=`docker exec -it lock-container /bin/sh -c "cat /etc/hostname"`

  log "Removing the current version of Lock for Container..."
  $DR docker stop $CONTAINER_NAME &> /dev/null
  $DR docker rm $CONTAINER_NAME &> /dev/null

  log "Installing the new version of Lock for Container..."

  install

  $DR docker cp $CONTAINER_NAME:/etc/tosibox/product ./tosibox-product-temporal-file &> /dev/null
  $DR docker cp $CONTAINER_NAME:/etc/tosibox/version ./tosibox-version-temporal-file &> /dev/null

  IMPORT_CONFIG=$CONFIG_FILENAME

  log "Importing configuration of Lock for Container..."
  import_config

  # copy the version / product filers of the new release to the new configs
  $DR docker cp ./tosibox-product-temporal-file $CONTAINER_NAME:/etc/tosibox/product
  $DR docker cp ./tosibox-version-temporal-file $CONTAINER_NAME:/etc/tosibox/version
  docker_exec "echo 'nameserver 8.8.8.8' >> /etc/resolv.conf"
  docker_exec "pkill -HUP -x tosiboxd"
  # Restart tbstatus to make sure interprocess communication works
  docker_exec "/etc/init.d/tbstatus restart &> /dev/null"

  rm ./tosibox-product-temporal-file ./tosibox-version-temporal-file &> /dev/null

  $DR docker network disconnect bridge $CONTAINER_NAME &> /dev/null

  $DR docker network connect $LAN_NW_NAME $CONTAINER_NAME &> /dev/null

  log "Restarting Lock for Container..."

  restart

  log "Update finished."
}

uninstall() {
  while true; do
    log
    log "This will remove all matched TOSIBOX® Keys and uninstall Lock for Container."
    log "All VPN connections will be closed. If you are remotely connected you will be disconnected."
    read -p "Do you want to uninstall Lock for Container? (y/n) " yn
    case $yn in
        [Yy]* ) break;;
        [Nn]* ) log "Cancelled."; exit;;
        * ) log "Please answer yes or no.";;
    esac
  done

  log "Reset all matched TOSIBOX® Keys..."
  docker_exec "/usr/share/tosibox/reset_pairings.sh -yes"
  log
  log "Uninstalling Lock for Container..."
  $DR docker stop $CONTAINER_NAME &> /dev/null
  $DR docker rm $CONTAINER_NAME &> /dev/null
  # Remove network configurations (if exist)...
  $DR docker network remove $LAN_NW_NAME &> /dev/null
  #$DR $sudo ip link del $LAN_LINK_NAME &> /dev/null  # if created for macvlan
  log
  log "Uninstall finished."
  log
  log "If you persisted network configurations during install, remove Lock for Container additions from /etc/network/interfaces and /etc/sysctl.conf files"
  log "If you don't need routing anymore, set /sbin/sysctl net.ipv4.conf.all.forwarding=0 and reset your IPTABLES FORWARDING rules."
}

remote_matching_start() {
  log "Starting remote remote matching..."
  log "Remote Matching Code for this Lock for Container:"
  log ""
  docker_exec "/usr/share/tosibox/remote-matching.lua start" "visible"
  log ""
  log "Use this code for remote matching your Key with this Lock for Container."
}

remote_matching_stop() {
  log "Cancelling remote matching..."
  docker_exec "/usr/share/tosibox/remote-matching.lua stop"
  log
  log "Remote Matching Code destroyed."
}

export_config() {
  log "Creating .tar export from the container..."
  CONFIG_FILENAME=$EXPORT_CONFIG"_config_"$(date "+%y%m%d%H%M%S").tar
  docker_exec "tar cfz $CONFIG_FILENAME /etc/shadow /etc/passwd /etc/config/ /etc/tosibox/ /etc/openvpn/ /etc/dropbear/ /usr/share/tosibox/authorized_keys &>/dev/null"
  $DR docker cp $CONTAINER_NAME:$CONFIG_FILENAME .
}

import_config() {
  log "Importing" $IMPORT_CONFIG
  $DR docker cp $IMPORT_CONFIG $CONTAINER_NAME:$IMPORT_CONFIG
  docker_exec "tar xfz "$IMPORT_CONFIG
}

install_wizard() {
  log
  log "Welcome to the TOSIBOX® Lock for Container installation wizard."
  log

  # Setup WebUI port
  log "======================================================="
  log "STEP 1: Lock for Container web user interface port."
  log "Which port should be used to access Lock for Container?"
  log "Type in the port and press <Enter>."
  log "======================================================="
  log

  while true; do
    read -p "Port (8000): " PORT
    if [ -z "$PORT" ]; then
      PORT="8000"
      break;
    elif [[ $PORT != *[!0-9]* ]]; then
      break;
    else
      log "Please specify valid number"
    fi
  done
  log

  # Install the Docker image
  install
  log

  log "======================================================="
  log "STEP 2: Configure networking and routing. Container "
  log "requires a network, and next a bridge network will be "
  log "created."
  log "If connected to TOSIBOX® Virtual Central Lock, each"
  log "Lock for Container should have different network space. "
  log "Please answer the next questions to setup networking."
  log "======================================================="


  # Add default Container subnet config
  if [ -f "$CONTAINER_IP_FILE" ]; then
    DEFAULT_SUBNET=$(cat $CONTAINER_IP_FILE)
  else
    DEFAULT_SUBNET="10.10.$(($RANDOM%255)).0"
    echo $DEFAULT_SUBNET > $CONTAINER_IP_FILE
    $DR docker cp $CONTAINER_IP_FILE $CONTAINER_NAME:/etc/config/
  fi
  DEFAULT_NETMASK="255.255.255.0"
  log
  log "A random network range $DEFAULT_SUBNET is generated for this Lock for Container instance. You may change it to something else."
  log "Ensure you don't have conflicting network ranges in your infrastructure."

  LAN_TYPE="internal"

  ask_network_config

  # Add default static route to $DR docker network
  log
  log "Next, configure routing to the Container subnet."
  log

  DEFAULT_NETWORK="$DEFAULT_SUBNET"

  ask_route_config

  IP_ADDRESSES=$(ifconfig | grep "inet " | grep -v "127.0.0.1" | sed -e $'s/[[:blank:]]*inet //g')

  # Add additional routes?
  DEFAULT_NETWORK=$(echo $IP_ADDRESSES | awk '{print $1;}' | cut -d ":" -f 2 | cut -d "." -f 1-3).0
  DEFAULT_NETMASK=$(echo $IP_ADDRESSES | awk '{print $3;}' | cut -d ":" -f 2 | sed -e $'s/0x//g')
  while true; do
    log
    log "You have the following IP addresses for the host. If you want traffic to be routed there add routing to these networks."
    log
    log "$IP_ADDRESSES"
    log
    read -p "Add routing to additional networks? (y/n): " yn
    case $yn in
        [Yy]* ) ask_route_config;;
        [Nn]* ) break;;
        * ) log "Please answer (y)es or (n)o.";;
    esac
  done
  log

  log "Restarting Lock for Container..."
  restart
  log
  log "======================================================="
  log "STEP 3: Activate Lock for Container."
  log "Do you want to activate Lock for Container now?"
  log "If you have an Activation Code you can activate"
  log "Lock for Container now or do it via web user interface"
  log "after installation."
  log "Type y or n and press <Enter>"
  log "======================================================="
  log
  while true; do
    read -p "Activate Lock for Container (y/n) " yn
    case $yn in
        [Yy]* ) ACTIVATE=1; break;;
        [Nn]* ) activation_help; break;;
        * ) log "Please answer (y)es or (n)o.";;
    esac
  done

  # Check for valid license key
  if [ $ACTIVATE ]; then
    ask_activation_license
  fi

  if [ $LICENSE_VALID -eq 1 ]; then
    activate
  fi

  # Setup variables for remote matching if Lock for Container was activated
  if [ $ACTIVATE ]; then
    remote_matching_start
    ACTIVATE=""
  fi
}

change_password() {
  log "Setup the admin password to login to the system via the web interface."
  read -p "Enter new admin password (Default: $DEFAULT_PASSWORD): " PASSWORD
  while true; do
    if [ $PASSWORD ]; then
      set_admin_passwd $PASSWORD
      break;
    elif [ -z $PASSWORD ] && [ $DEFAULT_PASSWORD ]; then
      set_admin_passwd $DEFAULT_PASSWORD
      break;
    fi
  done
}

set_admin_passwd() {
  PASSWORD="$1"

  docker_exec "echo -e '$PASSWORD\n$PASSWORD' | passwd admin &> /dev/null"
}

docker_exec() {
  COMMAND="$1"
  if [ "$DR" == "$DRYRUN_ECHO" ]; then
    $DR docker exec -it $CONTAINER_NAME /bin/sh -c "$COMMAND"
  elif [ "$2" == "visible" ]; then
    $DR docker exec -it $CONTAINER_NAME /bin/sh -c "$COMMAND"
  else
    $DR docker exec -it $CONTAINER_NAME /bin/sh -c "$COMMAND" &> /dev/null
  fi
}


################################################################################
# MAIN SCRIPT
################################################################################

if [ -z "$1" ]; then
  log "$HELPOPTIONS"
  exit 1
fi

while [[ "$#" -gt 0 ]]; do case $1 in
  install) INSTALL=1;;
  -m|--hostname) HOSTNAME="$2"; shift;;
  -p|--port) PORT="$2"; shift;;
  -r|--variant) VARIANT="$2"; shift;;
  --namespace) NAMESPACE="$2"; shift;;
  --wizard) WIZARD=1; shift;;
  activate) ACTIVATE=1;;
  -l|--license) LICENSE="$2"; shift;;
  uninstall) UNINSTALL=1;;
  start) START=1;;
  stop) STOP=1;;
  restart) RESTART=1;;
  update) UPDATE=1;;
  route) ROUTECONFIG=1;;
  -n|--network) NETWORK_CIDR="$2"; NETWORK=$(echo $NETWORK_CIDR | cut -d'/' -f1);shift;;
  -g|--gateway) GATEWAY="$2"; shift;;
  -a|--netmask) NETMASK="$2"; shift;;
  hostroute) HOSTROUTECONFIG=1;;
  --hr-network) HOSTROUTE_NETWORK_CIDR="$2"; HOSTROUTE_NETWORK=$(echo $HOSTROUTE_NETWORK_CIDR | cut -d'/' -f1);shift;;
  --hr-gateway) HOSTROUTE_GATEWAY="$2"; shift;;
  --hr-netmask) HOSTROUTE_NETMASK="$2"; shift;;
  net) NETCONFIG=1;;
  -y|--type) LAN_TYPE="$2"; shift;;
  -s|--subnet) LAN_SUBNET_CIDR="$2"; LAN_SUBNET_PREFIX=$(echo $LAN_SUBNET_CIDR | cut -d'/' -f2); shift;;
  -w|--langw) LAN_GW="$2"; shift;;
  -t|--host) LAN_HOST_ADDR="$2"; shift;;
  -i|--interface) LAN_INTERFACE="$2"; shift;;
  -k|--lock) LAN_LOCK_ADDR="$2"; shift;;
  --persist) PERSIST_NETCONFIG=1;;
  shell) OPEN_SHELL=1;;
  match) MATCH=1;;
  --start) REMOTE_MATCHING_START=1;;
  --cancel) REMOTE_MATCHING_CANCEL=1;;
  --reset) RESET_MATCHED=1;;
  backup) BACKUP=1;;
  --import-config) IMPORT_CONFIG="$2"; shift;;
  --export-config) EXPORT_CONFIG="$2"; shift;;
  -b|--import) IMPORT_FILE="$2"; shift;;
  -x|--export) EXPORT_FILE="$2"; shift;;

  -h|--help) echo "$OPTIONS";exit 0; shift;;
  -v|--version) echo "$VERSION";exit 0; shift;;
  --dryrun) DR=$DRYRUN_ECHO; shift;;

  *) echo "Unknown parameter passed: $1"; exit 1;;
esac; shift;
done

if [ "$DR" == "$DRYRUN_ECHO" ]; then
  log
  log RUNNING IN DRYRUN MODE. This will not cause changes to the system.
  $DR commands run in dryrun mode are shown in red.
  log Commands that output to /dev/null may not be visible.
  # wizards check that there's only one parameter, but we want to run wizards also in dryrun mode
  PARAMCOUNT=$(($PARAMCOUNT-1));
fi

if [ $NAMESPACE ]; then
  LAN_NW_NAME="$LAN_NW_NAME-$NAMESPACE"
  CONTAINER_NAME="$CONTAINER_NAME-$NAMESPACE"
  LAN_LINK_NAME="$LAN_LINK_NAME-$NAMESPACE"
  CONTAINER_IP_FILE="$CONTAINER_IP_FILE-$NAMESPACE"

  # wizards check that there's only one parameter, but we want to run wizards also in namespace mode
  PARAMCOUNT=$(($PARAMCOUNT-2));
fi

# Use sudo if not running as root
if [ "$EUID" -ne 0 ]; then
  sudo="sudo"
fi

if [ $NETCONFIG ] && [ $PARAMCOUNT -lt 2 ]; then
  ask_network_config
fi

if [ $NETCONFIG ] && [ "$LAN_SUBNET_CIDR" ]; then
  add_network_config
fi

if [ $INSTALL ]; then
  if [ $PARAMCOUNT -lt 2 ]; then
    install_wizard
  else
    install
  fi

  change_password

  log
  log "======================================================="
  log "INSTALLATION READY!"
  log
  if [ -z $ACTIVATE ]; then
    log "You can now use Lock for Container via web user"
    log "interface at http://localhost:$PORT."
    log "To add Lock for Container to your TOSIBOX(R) network"
    log "use the Remote Matching functionality."
    log "You can find Remote Matching at"
    log "   Settings -> Keys and Locks"
    log "in the Lock for Container web user interface."
  else
    log "You can now activate Lock for Container via the"
    log "web user interface at http://localhost:$PORT."
  fi
  log
  log "If you want Lock for Container to route traffic to "
  log "networks it has access to from the host machine"
  log "(e.g. through VCL) do the following:"
  log "# route add -net [your target network] \\"
  log "   netmask 255.255.0.0 gw $LAN_GW"
  log "or:"
  log "# ip route add [your target network] via $LAN_GW"
  log "======================================================="
fi

if [ $ROUTECONFIG ] && [ $NETWORK ] && [ $NETMASK ]; then
  add_static_route
fi

if [ $ROUTECONFIG ] && [ $PARAMCOUNT -lt 2 ]; then
  ask_route_config
fi

if [ $HOSTROUTECONFIG ] && [ $PARAMCOUNT -lt 2 ]; then
  ask_hostroute_config
fi

if [ $HOSTROUTECONFIG ] && [ $HOSTROUTE_NETWORK ] && [ $HOSTROUTE_GATEWAY ]; then
  add_hostroute
fi

# if macvlan network type is used, need to configure also ip settings of the host
if [ $NETCONFIG ] && [ $LAN_TYPE = "external" ] && [ $LAN_SUBNET_CIDR ] && [ $LAN_INTERFACE ] && [ $LAN_LINK_NAME ]; then
  log "Configuring links and routes... sudo privileges required."
  $DR $sudo ip link delete $LAN_LINK_NAME
  # Create a link for macvlan
  $DR $sudo ip link add $LAN_LINK_NAME link $LAN_INTERFACE type macvlan mode bridge
  # Set an ip address for the device in macvlan subnet
  $DR $sudo ip addr add $LAN_HOST_ADDR dev $LAN_LINK_NAME
  $DR $sudo ifconfig $LAN_LINK_NAME up
  $DR $sudo ip addr add $LAN_HOST_ADDR/$LAN_SUBNET_PREFIX dev $LAN_LINK_NAME

  # Wait a while so interfaces have time to wake up
  $DR wait_until_ready
  docker_exec "ip route del default via $LAN_GW dev br-lan"
  docker_exec "ip route add default via $LAN_HOST_ADDR dev br-lan"
fi

if [ $START ]; then
  docker start $CONTAINER_NAME
fi

if [ $STOP ]; then
  stop
fi

if [ $RESTART ]; then
  restart
fi

if [ $UNINSTALL ]; then
  uninstall
fi

if [[ -n "$ACTIVATE" ]] && [[ -n "$LICENSE" ]]; then
  activate
elif [ $ACTIVATE ]; then
  ask_activation_license
  if [ "$LICENSE_VALID" -eq 1 ]; then
    activate
  fi
fi

if [ $BACKUP ] && [ $EXPORT_FILE ]; then
  log "Creating .tar export from the container..."
  $DR docker export $CONTAINER_NAME > $EXPORT_FILE"_"$(date "+%y%m%d%H%M%S").tar
fi

if [ $BACKUP ] && [ "$IMPORT_FILE" ]; then
  log "Importing $IMPORT_FILE"
  cat "$IMPORT_FILE" | docker import - $CONTAINER_NAME:$VARIANT
  $DR docker run -itd --network=$LAN_NW_NAME --restart=unless-stopped --name $CONTAINER_NAME -p $PORT:80 --cap-add=NET_ADMIN --cap-add=SYS_TIME --cap-add=SYS_PTRACE $CONTAINER_NAME:$VARIANT /sbin/init
fi

if [ $BACKUP ] && [ $EXPORT_CONFIG ]; then
  export_config
fi

if [ $BACKUP ] && [ "$IMPORT_CONFIG" ]; then
  import_config
fi


if [ $OPEN_SHELL ]; then
  log "Opening shell to Lock for Container.."
  $DR docker exec -it $CONTAINER_NAME /bin/sh
  exit 0
fi

if [ $MATCH ] && [ $REMOTE_MATCHING_START ]; then
  remote_matching_start
fi

if [ $MATCH ] && [ $REMOTE_MATCHING_CANCEL ]; then
  remote_matching_stop
fi

if [ $MATCH ] && [ $RESET_MATCHED ]; then
  log "Reset all matched TOSIBOX® Keys..."
  docker_exec "/usr/share/tosibox/reset_pairings.sh -yes"
  log
  log "All matched TOSIBOX® Keys reset."
fi

if [ $ACTIVATE ] && [ $LICENSE ]; then
  log "You may now open a browser in http://localhost:$PORT to access the web user interface."
fi

if [ $UPDATE ]; then
  update
fi

if [ $CHANGE_PASSWORD ]; then
  change_password
fi

exit 0
